# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from earth2studio.data import DataSource
from earth2studio.data.utils import fetch_data
from earth2studio.statistics.utils import _broadcast_weights
from earth2studio.utils.coords import handshake_coords, handshake_dim
from earth2studio.utils.type import CoordSystem


class acc:
    """
    Statistic for calculating the anomaly correlation coefficient of two tensors
    over a set of given dimensions, with respect to some optional climatology.

    Parameters
    ----------
    reduction_dimensions: List[str]
        A list of names corresponding to dimensions to perform the
        statistical reduction over. Example: ['lat', 'lon']
    climatology: DataSource
        Optional (by default None) climatology to remove from tensors to create
        anomalies before computing the correlation coefficient.
    weights: torch.Tensor = None
        A tensor containing weights to assign to the reduction dimensions.
        Note that these weights must have the same number of dimensions
        as passed in reduction_dimensions.
        Example: if reduction_dimensions = ['lat', 'lon'] then
        assert weights.ndim == 2.
    """

    def __init__(
        self,
        reduction_dimensions: list[str],
        climatology: DataSource | None = None,
        weights: torch.Tensor | None = None,
    ):

        if weights is not None:
            if weights.ndim != len(reduction_dimensions):
                raise ValueError(
                    "Error! Weights must be the same dimension as reduction_dimensions"
                )
        self._reduction_dimensions = reduction_dimensions
        self.weights = weights

        self.climatology = climatology

    def __str__(self) -> str:
        return "_".join(self._reduction_dimensions + ["acc"])

    @property
    def reduction_dimensions(self) -> list[str]:
        return self._reduction_dimensions

    def output_coords(self, input_coords: CoordSystem) -> CoordSystem:
        """Output coordinate system of the computed statistic, corresponding to the given input coordinates

        Parameters
        ----------
        input_coords : CoordSystem
            Input coordinate system to transform into output_coords

        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """
        output_coords = input_coords.copy()
        for dimension in self.reduction_dimensions:
            handshake_dim(input_coords, dimension)
            output_coords.pop(dimension)

        return output_coords

    def __call__(
        self,
        x: torch.Tensor,
        x_coords: CoordSystem,
        y: torch.Tensor,
        y_coords: CoordSystem,
    ) -> tuple[torch.Tensor, CoordSystem]:
        """
        Apply metric to data `x` and `y`, checking that their coordinates
        are broadcastable.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor, typically the forecast or prediction tensor, but ACC is
            symmetric with respect to `x` and `y`.
        x_coords : CoordSystem
            Ordered dict representing coordinate system that describes the `x` tensor.
            `reduction_dimensions` must be in coords.
            "time" and "variable" must be in x_coords.
        y : torch.Tensor
            Input tensor #2 intended to be used as validation data, but ACC is symmetric
            with respect to `x` and `y`.
        y_coords : CoordSystem
            Ordered dict representing coordinate system that describes the `y` tensor.
            `reduction_dimensions` must be in coords.
            "time" and "variable" must be in y_coords.
            If "lead_time" is in x_coords, then "lead_time" must also be in y_coords. The
            intention, in this case, is that users will use `fetch_data` to make it easier
            to match validation times.

        Returns
        -------
        tuple[torch.Tensor, CoordSystem]
            Returns anomaly correlation coefficient tensor with appropriate reduced coordinates.

        Note
        ----
        Reference: https://www.atmos.albany.edu/daes/atmclasses/atm401/spring_2016/ppts_pdfs/ECMWF_ACC_definition.pdf
        """

        # Input coordinate checking
        handshake_dim(x_coords, "time")
        handshake_dim(x_coords, "variable")
        for i, c in enumerate(x_coords):
            handshake_dim(y_coords, c, i)
            handshake_coords(x_coords, y_coords, c)

        dims = [list(x_coords).index(rd) for rd in self._reduction_dimensions]
        output_coords = self.output_coords(x_coords)

        weights = _broadcast_weights(
            self.weights, self._reduction_dimensions, x_coords
        ).to(x.device)
        weights_sum = torch.sum(weights)

        # Get climatology information
        if self.climatology is not None:
            if "lead_time" in output_coords:
                clim, _ = fetch_data(
                    self.climatology,
                    output_coords["time"],
                    output_coords["variable"],
                    lead_time=output_coords["lead_time"],
                    device=x.device,
                )
            else:
                da = self.climatology(output_coords["time"], output_coords["variable"])
                clim = torch.as_tensor(da.values, device=x.device, dtype=x.dtype)
        else:
            clim = torch.zeros_like(x)

        x_hat = x - clim
        x_bar = torch.sum(weights * x_hat, dim=dims, keepdim=True) / weights_sum
        x_diff = x_hat - x_bar

        y_hat = y - clim
        y_bar = torch.sum(weights * y_hat, dim=dims, keepdim=True) / weights_sum
        y_diff = y_hat - y_bar

        p1 = torch.sum(weights * x_diff * y_diff, dim=dims)
        p2 = torch.sum(weights * x_diff * x_diff, dim=dims)
        p3 = torch.sum(weights * y_diff * y_diff, dim=dims)

        return p1 / torch.sqrt(p2 * p3), output_coords



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import http.client
import os
import re
import urllib.parse
from typing import Any

import aiohttp
from fsspec.asyn import sync_wrapper
from fsspec.callbacks import DEFAULT_CALLBACK
from fsspec.implementations.http import HTTPFileSystem
from ngcbpc.api import utils as rest_utils
from ngcbpc.api.authentication import Authentication
from ngcbpc.api.configuration import Configuration
from ngcbpc.util.utils import format_org_team


async def get_client(**kwargs) -> aiohttp.ClientSession:  # type: ignore
    """Default aiohttp client"""
    return aiohttp.ClientSession(**kwargs)


class NGCModelFileSystem(HTTPFileSystem):
    """NGC model registry file system for pulling and opening files that are stored
    on both public and private model registries. Largely a wrapper ontop of fsspec
    HTTPFileSystem class. This works using a url with the following pattern:

    `ngc://models/<org_id/team_id/model_id>@<version>/<path/in/repo>`

    or if no team

    `ngc://models/<org_id/model_id>@<version>/<path/in/repo>`

    For example:
    `ngc://models/nvidia/modulus/sfno_73ch_small@0.1.0/sfno_73ch_small/config.json`

    Note
    ----
    For private registries use one of the standard authentication methods such as
    an environment variable (NGC_CLI_API_KEY) or ngc config file a NGC API key that has
    access to the respective model registry.

    - https://docs.ngc.nvidia.com/cli/script.html?highlight=ngc_cli_api
    - https://docs.ngc.nvidia.com/cli/cmd.html

    Parameters
    ----------
    block_size: int
        Blocks to read bytes; if 0, will default to raw requests file-like
        objects instead of HTTPFile instances
    simple_links: bool
        If True, will consider both HTML <a> tags and anything that looks
        like a URL; if False, will consider only the former.
    same_scheme: bool
        When doing ls/glob, if this is True, only consider paths that have
        http/https matching the input URLs.
    size_policy: this argument is deprecated
    client_kwargs: dict
        Passed to aiohttp.ClientSession, see
        https://docs.aiohttp.org/en/stable/client_reference.html
        For example, ``{'auth': aiohttp.BasicAuth('user', 'pass')}``
    get_client: Callable[..., aiohttp.ClientSession]
        A callable which takes keyword arguments and constructs
        an aiohttp.ClientSession. It's state will be managed by
        the HTTPFileSystem class.
    storage_options: key-value
        Any other parameters passed on to requests
    cache_type, cache_options: defaults used in open
    """

    def __init__(  # type: ignore
        self,
        block_size=None,
        simple_links=True,
        same_scheme=True,
        size_policy=None,
        cache_type="bytes",
        cache_options=None,
        asynchronous=False,
        loop=None,
        client_kwargs=None,
        get_client=get_client,
        encoded=False,
        **storage_options,
    ):
        config = Configuration()
        config._sdk_configuration.db = True
        config._load_from_env_vars()
        Authentication.config = config

        super().__init__(
            simple_links,
            block_size,
            same_scheme,
            size_policy,
            cache_type,
            cache_options,
            asynchronous,
            loop,
            client_kwargs,
            get_client,
            encoded,
            **storage_options,
        )
        # Bit of a hack to easily get a sync version of this function
        # https://github.com/fsspec/filesystem_spec/blob/master/fsspec/asyn.py#L275
        # open requires the creation of a HttpFile, easier to sync intercept this at the start
        mth = sync_wrapper(getattr(self, "_get_model_asset_url"), obj=self)
        setattr(self, "get_model_asset_url", mth)

    def _parse_ngc_uri(
        self, root: str
    ) -> tuple[str, str, str | None, str | None, str | None]:
        """Parse NGC url into components"""
        # ngc://models/<org_id/team_id/model_id>@<version>/<path/in/repo>`
        suffix = "ngc://models/"
        # The regex check
        pattern = re.compile(rf"{suffix}[\w-]+(/[\w-]+)?(/[\w-]+)?@[A-Za-z0-9.]+")
        if not pattern.match(root):
            raise ValueError(
                "Invalid URL, should be of form ngc://models/<org_id/team_id/model_id>@<version>\n"
                + f" got {root}"
            )
        root = root.replace(suffix, "")
        if len(root.split("@")[0].split("/")) == 3:
            (org, team, model_info) = root.split("/", 2)
            (name, model_info) = model_info.split("@", 1)
        elif len(root.split("@")[0].split("/")) == 2:
            (org, model_info) = root.split("/", 1)
            (name, model_info) = model_info.split("@", 1)
            team = None
        else:
            (name, model_info) = root.split("@", 1)
            org = None
            team = None

        # Extract filename / path if present
        if len(model_info.rstrip("/").split("/")) > 1:
            (version, filepath) = model_info.split("/", 1)
        else:
            version = model_info.rstrip("/")
            filepath = None

        return name, version, org, team, filepath

    def _get_ngc_model_url(
        self,
        name: str,
        version: str,
        org: str = None,
        team: str = None,
        filepath: str = None,
        authenticated_api: bool = True,
    ) -> str:
        # Authenticated API
        if authenticated_api:
            url = "https://api.ngc.nvidia.com/v2/"
            if format_org_team(org, team):
                url += f"{format_org_team(org, team)}/"
            url += f"models/{name}/{version}/files"
            if filepath:
                url = f"{url}?path={filepath}"
        # Public API
        else:
            url = "https://api.ngc.nvidia.com/v2/models/"
            relative_urls = []
            if org:
                relative_urls += [org]
            if team:
                relative_urls += [team]
            relative_urls += [name, "versions", version, "files/"]
            url = urllib.parse.urljoin(url, os.path.join(*relative_urls))
            if filepath:
                url = urllib.parse.urljoin(url, filepath)
        return url

    async def _get_ngc(
        self, asset_url: str, headers: dict[str, str] = {}
    ) -> tuple[int, dict[str, Any] | None]:
        """Send get request to NGC and recieves asset reponse

        Parameters
        ----------
        url : str
            File/asset URL
        headers :  dict[str, str], optional
            Headers for AuthN GET request, by default {}

        Returns
        -------
        tuple[int, str]
            Status code and response JSON
        """
        # Based on: ngc/apps/ngc-cli/-/blob/main/ngcbpc/ngcbpc/transfer/async_download.py?ref_type=heads#L782
        # Get the direct download URL
        async with aiohttp.ClientSession() as session:
            # Use http filesystem encode URL
            dl_url_resp = await session.get(self.encode_url(asset_url), headers=headers)
            status = dl_url_resp.status
            if status == http.client.OK:
                direct_url_dict = await dl_url_resp.json()
                return status, direct_url_dict

            return status, None

    async def _get_model_asset_url(self, rpath: str) -> str:
        # Check if ngc url, if not just pass
        if not rpath.startswith("ngc://"):
            return rpath
        # Parse remote url
        name, version, org, team, filepath = self._parse_ngc_uri(rpath)
        # Create headers to determine if we have authn headers and point to private vs public APIs
        auth_header = Authentication.auth_header(auth_org=org, auth_team=team)
        headers = rest_utils.default_headers(auth_header)

        direct_url = None
        if "Authorization" in headers:
            api_type = "private"
            url = self._get_ngc_model_url(name, version, org, team, filepath, True)
            status, response = await self._get_ngc(url, headers)
            # Attempt authn with renew
            if status == http.client.UNAUTHORIZED:
                auth_header = Authentication.auth_header(
                    auth_org=org, auth_team=team, renew=True
                )
                headers = rest_utils.default_headers(auth_header)
                status, response = await self._get_ngc(url, headers)

            if status == http.client.OK and response is not None:
                direct_url = response["urls"][0]
        # No API headers created, so fall back to public access method
        else:
            api_type = "public"
            # Check to see if asset is there on NGC at all
            url = self._get_ngc_model_url(name, version, org, team, None, False)
            status, response = await self._get_ngc(url)
            if status == http.client.OK and response is not None:
                paths = [
                    file["path"] for file in response["modelFiles"]
                ]  # List of all model files
                if filepath in paths:
                    direct_url = self._get_ngc_model_url(
                        name, version, org, team, filepath, False
                    )
                else:
                    status = 404

        #  Do some graceful error catching down here...
        if status == http.client.UNAUTHORIZED:
            raise http.client.HTTPException(
                f"Unauthorized NGC API key for {api_type} model package"
            )
        if status == http.client.NOT_FOUND:
            raise http.client.HTTPException(
                f"Requested {api_type} model package {rpath} not found"
            )
        if not direct_url:
            raise http.client.HTTPException(
                f"Failed to get valid download URL for {api_type} model package"
            )

        return direct_url

    async def _expand_path(self, path, recursive=False, maxdepth=None):  # type: ignore
        # hhttps://github.com/fsspec/filesystem_spec/blob/master/fsspec/asyn.py#L864
        # TODO: Improve to allow download of multiple files
        # get -> expand_path -> ``glob`` or ``find``, which may in turn call ``ls``
        # To find possible files one needs to hit the /file/ root and pull out
        # possible URLS from the response
        # E.g. https://api.ngc.nvidia.com/v2/models/nvidia/modulus/sfno_73ch_small/versions/0.1.0/files/
        # could add pages via apps/ngc-cli/-/blob/main/ngcbpc/ngcbpc/transfer/async_download.py?ref_type=heads#L817
        # response['modelFiles'] lists possible files
        raise NotImplementedError("Glob / recursive patterns not supported yet")
        return await super()._expand_path(path, recursive, maxdepth)

    async def _glob(self, path, maxdepth=None, **kwargs):  # type: ignore
        raise NotImplementedError("Glob / recursive patterns not supported yet")
        return await super()._glob(path, maxdepth, **kwargs)

    async def _find(self, path, maxdepth=None, withdirs=False, **kwargs):  # type: ignore
        raise NotImplementedError("Glob / recursive patterns not supported yet")
        return await super()._find(path, maxdepth, withdirs, **kwargs)

    async def _ls_real(self, url, detail=True, **kwargs):  # type: ignore
        url = await self._get_model_asset_url(url)
        return await super()._ls_real(url, detail, **kwargs)

    async def _get_file(  # type: ignore
        self, rpath, lpath, chunk_size=5 * 2**20, callback=DEFAULT_CALLBACK, **kwargs
    ):
        # Calling .get will call get_file via the AbstractFileSystem
        # https://github.com/fsspec/filesystem_spec/blob/master/fsspec/spec.py#L916
        # Will get converted to sync function via Metaclass
        # https://github.com/fsspec/filesystem_spec/blob/master/fsspec/asyn.py#L931
        url = await self._get_model_asset_url(rpath)
        return await super()._get_file(
            url, lpath, chunk_size=chunk_size, callback=callback, **kwargs
        )

    def _open(  # type: ignore
        self,
        path,
        mode="rb",
        block_size=None,
        autocommit=None,  # XXX: This differs from the base class.
        cache_type=None,
        cache_options=None,
        size=None,
        **kwargs,
    ):
        path = self.get_model_asset_url(path)
        return super()._open(
            path,
            mode,
            block_size,
            autocommit,
            cache_type,
            cache_options,
            size,
            **kwargs,
        )



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from earth2studio.utils import handshake_dim
from earth2studio.utils.type import CoordSystem


class Brown:
    """Lat/Lon 2D brown noise

    Parameters
    ----------
    noise_amplitude : float | Tensor, optional
        Noise amplitude, by default 0.05. If a tensor,
        this must be broadcastable with the input data.
    reddening : int, optional
        Reddening in Fourier space, by default 2
    """

    def __init__(
        self, noise_amplitude: float | torch.Tensor = 0.05, reddening: int = 2
    ):
        self.reddening = reddening
        self.noise_amplitude = (
            noise_amplitude
            if isinstance(noise_amplitude, torch.Tensor)
            else torch.Tensor([noise_amplitude])
        )

    @torch.inference_mode()
    def __call__(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Apply perturbation method

        Parameters
        ----------
        x : torch.Tensor
            Input tensor intended to apply perturbation on
        coords : CoordSystem
            Ordered dict representing coordinate system that describes the tensor, must
            contain "lat" and "lon" coordinates in last two dims

        Returns
        -------
        tuple[torch.Tensor, CoordSystem]:
            Output tensor and respective coordinate system dictionary
        """
        shape = x.shape
        # Check the required dimensions are present
        handshake_dim(coords, required_dim="lat", required_index=-2)
        handshake_dim(coords, required_dim="lon", required_index=-1)

        noise = self._generate_noise_correlated(tuple(shape), device=x.device)
        noise_amplitude = self.noise_amplitude.to(x.device)
        return x + noise_amplitude * noise, coords

    def _generate_noise_correlated(
        self, shape: tuple[int, ...], device: torch.device
    ) -> torch.Tensor:
        """Utility class for producing brown noise."""
        noise = torch.randn(*shape, device=device)
        x_white = torch.fft.rfft2(noise)
        S = (
            torch.abs(torch.fft.fftfreq(shape[-2], device=device).reshape(-1, 1))
            ** self.reddening
            + torch.fft.rfftfreq(shape[-1], device=device) ** self.reddening
        )
        S = 1 / S
        S[..., 0, 0] = 0
        S = S / torch.sqrt(torch.mean(S**2))

        x_shaped = x_white * S
        noise_shaped = torch.fft.irfft2(x_shaped, s=shape[-2:])
        return noise_shaped



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable

import numpy as np

from .base import LexiconType


class GEFSLexicon(metaclass=LexiconType):
    """Global Ensemble Forecast System Lexicon, right now only support isobarric
    GEFS vocab specified <Data class>::<Parameter ID>::<Level/ Layer>

    Note
    ----
    Additional resources:
    - https://www.nco.ncep.noaa.gov/pmb/products/gens/gec00.t00z.pgrb2a.0p50.f003.shtml
    - https://www.nco.ncep.noaa.gov/pmb/products/gens/gec00.t00z.pgrb2b.0p50.f003.shtml
    """

    VOCAB = {
        "u10m": "pgrb2a::UGRD::10 m above ground",
        "v10m": "pgrb2a::VGRD::10 m above ground",
        "u100m": "pgrb2b::UGRD::100 m above ground",
        "v100m": "pgrb2b::VGRD::100 m above ground",
        "t2m": "pgrb2a::TMP::2 m above ground",
        "d2m": "pgrb2b::DPT::2 m above ground",
        "r2m": "pgrb2a::RH::2 m above ground",
        "q2m": "pgrb2b::SPFH::2 m above ground",
        "t100m": "pgrb2b::TMP::100 m above ground",
        "sp": "pgrb2a::PRES::surface",
        "msl": "pgrb2b::PRES::mean sea level",
        "tcwv": "pgrb2a::PWAT::entire atmosphere (considered as a single layer)",
        "u1": "pgrb2b::UGRD::1 mb",
        "u2": "pgrb2b::UGRD::2 mb",
        "u3": "pgrb2b::UGRD::3 mb",
        "u5": "pgrb2b::UGRD::5 mb",
        "u7": "pgrb2b::UGRD::7 mb",
        "u10": "pgrb2a::UGRD::10 mb",
        "u20": "pgrb2b::UGRD::20 mb",
        "u30": "pgrb2b::UGRD::30 mb",
        "u50": "pgrb2a::UGRD::50 mb",
        "u70": "pgrb2b::UGRD::70 mb",
        "u100": "pgrb2a::UGRD::100 mb",
        "u150": "pgrb2b::UGRD::150 mb",
        "u200": "pgrb2a::UGRD::200 mb",
        "u250": "pgrb2a::UGRD::250 mb",
        "u300": "pgrb2a::UGRD::300 mb",
        "u350": "pgrb2b::UGRD::350 mb",
        "u400": "pgrb2a::UGRD::400 mb",
        "u450": "pgrb2b::UGRD::450 mb",
        "u500": "pgrb2a::UGRD::500 mb",
        "u550": "pgrb2b::UGRD::550 mb",
        "u600": "pgrb2b::UGRD::600 mb",
        "u650": "pgrb2b::UGRD::650 mb",
        "u700": "pgrb2a::UGRD::700 mb",
        "u750": "pgrb2b::UGRD::750 mb",
        "u800": "pgrb2b::UGRD::800 mb",
        "u850": "pgrb2a::UGRD::850 mb",
        "u900": "pgrb2b::UGRD::900 mb",
        "u925": "pgrb2a::UGRD::925 mb",
        "u950": "pgrb2b::UGRD::950 mb",
        "u975": "pgrb2b::UGRD::975 mb",
        "u1000": "pgrb2a::UGRD::1000 mb",
        "v1": "pgrb2b::VGRD::1 mb",
        "v2": "pgrb2b::VGRD::2 mb",
        "v3": "pgrb2b::VGRD::3 mb",
        "v5": "pgrb2b::VGRD::5 mb",
        "v7": "pgrb2b::VGRD::7 mb",
        "v10": "pgrb2a::VGRD::10 mb",
        "v20": "pgrb2b::VGRD::20 mb",
        "v30": "pgrb2b::VGRD::30 mb",
        "v50": "pgrb2a::VGRD::50 mb",
        "v70": "pgrb2b::VGRD::70 mb",
        "v100": "pgrb2a::VGRD::100 mb",
        "v150": "pgrb2b::VGRD::150 mb",
        "v200": "pgrb2a::VGRD::200 mb",
        "v250": "pgrb2a::VGRD::250 mb",
        "v300": "pgrb2a::VGRD::300 mb",
        "v350": "pgrb2b::VGRD::350 mb",
        "v400": "pgrb2a::VGRD::400 mb",
        "v450": "pgrb2b::VGRD::450 mb",
        "v500": "pgrb2a::VGRD::500 mb",
        "v550": "pgrb2b::VGRD::550 mb",
        "v600": "pgrb2b::VGRD::600 mb",
        "v650": "pgrb2b::VGRD::650 mb",
        "v700": "pgrb2a::VGRD::700 mb",
        "v750": "pgrb2b::VGRD::750 mb",
        "v800": "pgrb2b::VGRD::800 mb",
        "v850": "pgrb2a::VGRD::850 mb",
        "v900": "pgrb2b::VGRD::900 mb",
        "v925": "pgrb2a::VGRD::925 mb",
        "v950": "pgrb2b::VGRD::950 mb",
        "v975": "pgrb2b::VGRD::975 mb",
        "v1000": "pgrb2a::VGRD::1000 mb",
        "z1": "pgrb2b::HGT::1 mb",
        "z2": "pgrb2b::HGT::2 mb",
        "z3": "pgrb2b::HGT::3 mb",
        "z5": "pgrb2b::HGT::5 mb",
        "z7": "pgrb2b::HGT::7 mb",
        "z10": "pgrb2a::HGT::10 mb",
        "z20": "pgrb2b::HGT::20 mb",
        "z30": "pgrb2b::HGT::30 mb",
        "z50": "pgrb2a::HGT::50 mb",
        "z70": "pgrb2b::HGT::70 mb",
        "z100": "pgrb2a::HGT::100 mb",
        "z150": "pgrb2b::HGT::150 mb",
        "z200": "pgrb2a::HGT::200 mb",
        "z250": "pgrb2a::HGT::250 mb",
        "z300": "pgrb2a::HGT::300 mb",
        "z350": "pgrb2b::HGT::350 mb",
        "z400": "pgrb2b::HGT::400 mb",
        "z450": "pgrb2b::HGT::450 mb",
        "z500": "pgrb2a::HGT::500 mb",
        "z550": "pgrb2b::HGT::550 mb",
        "z600": "pgrb2b::HGT::600 mb",
        "z650": "pgrb2b::HGT::650 mb",
        "z700": "pgrb2a::HGT::700 mb",
        "z750": "pgrb2b::HGT::750 mb",
        "z800": "pgrb2b::HGT::800 mb",
        "z850": "pgrb2a::HGT::850 mb",
        "z900": "pgrb2b::HGT::900 mb",
        "z925": "pgrb2a::HGT::925 mb",
        "z950": "pgrb2b::HGT::950 mb",
        "z975": "pgrb2b::HGT::975 mb",
        "z1000": "pgrb2a::HGT::1000 mb",
        "t1": "pgrb2b::TMP::1 mb",
        "t2": "pgrb2b::TMP::2 mb",
        "t3": "pgrb2b::TMP::3 mb",
        "t5": "pgrb2b::TMP::5 mb",
        "t7": "pgrb2b::TMP::7 mb",
        "t10": "pgrb2a::TMP::10 mb",
        "t20": "pgrb2b::TMP::20 mb",
        "t30": "pgrb2b::TMP::30 mb",
        "t50": "pgrb2a::TMP::50 mb",
        "t70": "pgrb2b::TMP::70 mb",
        "t100": "pgrb2a::TMP::100 mb",
        "t150": "pgrb2b::TMP::150 mb",
        "t200": "pgrb2a::TMP::200 mb",
        "t250": "pgrb2a::TMP::250 mb",
        "t300": "pgrb2b::TMP::300 mb",
        "t350": "pgrb2b::TMP::350 mb",
        "t400": "pgrb2b::TMP::400 mb",
        "t450": "pgrb2b::TMP::450 mb",
        "t500": "pgrb2a::TMP::500 mb",
        "t550": "pgrb2b::TMP::550 mb",
        "t600": "pgrb2b::TMP::600 mb",
        "t650": "pgrb2b::TMP::650 mb",
        "t700": "pgrb2a::TMP::700 mb",
        "t750": "pgrb2b::TMP::750 mb",
        "t800": "pgrb2b::TMP::800 mb",
        "t850": "pgrb2a::TMP::850 mb",
        "t900": "pgrb2b::TMP::900 mb",
        "t925": "pgrb2a::TMP::925 mb",
        "t950": "pgrb2b::TMP::950 mb",
        "t975": "pgrb2b::TMP::975 mb",
        "t1000": "pgrb2a::TMP::1000 mb",
        "r10": "pgrb2a::RH::10 mb",
        "r20": "pgrb2b::RH::20 mb",
        "r30": "pgrb2b::RH::30 mb",
        "r50": "pgrb2a::RH::50 mb",
        "r70": "pgrb2b::RH::70 mb",
        "r100": "pgrb2a::RH::100 mb",
        "r150": "pgrb2b::RH::150 mb",
        "r200": "pgrb2a::RH::200 mb",
        "r250": "pgrb2a::RH::250 mb",
        "r300": "pgrb2b::RH::300 mb",
        "r350": "pgrb2b::RH::350 mb",
        "r400": "pgrb2b::RH::400 mb",
        "r450": "pgrb2b::RH::450 mb",
        "r500": "pgrb2a::RH::500 mb",
        "r550": "pgrb2b::RH::550 mb",
        "r600": "pgrb2b::RH::600 mb",
        "r650": "pgrb2b::RH::650 mb",
        "r700": "pgrb2a::RH::700 mb",
        "r750": "pgrb2b::RH::750 mb",
        "r800": "pgrb2b::RH::800 mb",
        "r850": "pgrb2a::RH::850 mb",
        "r900": "pgrb2b::RH::900 mb",
        "r925": "pgrb2a::RH::925 mb",
        "r950": "pgrb2b::RH::950 mb",
        "r975": "pgrb2b::RH::975 mb",
        "r1000": "pgrb2a::RH::1000 mb",
        "q10": "pgrb2b::SPFH::10 mb",
        "q20": "pgrb2b::SPFH::20 mb",
        "q30": "pgrb2b::SPFH::30 mb",
        "q50": "pgrb2b::SPFH::50 mb",
        "q70": "pgrb2b::SPFH::70 mb",
        "q100": "pgrb2b::SPFH::100 mb",
        "q150": "pgrb2b::SPFH::150 mb",
        "q200": "pgrb2b::SPFH::200 mb",
        "q250": "pgrb2b::SPFH::250 mb",
        "q300": "pgrb2b::SPFH::300 mb",
        "q350": "pgrb2b::SPFH::350 mb",
        "q400": "pgrb2b::SPFH::400 mb",
        "q450": "pgrb2b::SPFH::450 mb",
        "q500": "pgrb2b::SPFH::500 mb",
        "q550": "pgrb2b::SPFH::550 mb",
        "q600": "pgrb2b::SPFH::600 mb",
        "q650": "pgrb2b::SPFH::650 mb",
        "q700": "pgrb2b::SPFH::700 mb",
        "q750": "pgrb2b::SPFH::750 mb",
        "q800": "pgrb2b::SPFH::800 mb",
        "q850": "pgrb2b::SPFH::850 mb",
        "q900": "pgrb2b::SPFH::900 mb",
        "q925": "pgrb2b::SPFH::925 mb",
        "q950": "pgrb2b::SPFH::950 mb",
        "q975": "pgrb2b::SPFH::975 mb",
        "q1000": "pgrb2b::SPFH::1000 mb",
    }

    @classmethod
    def get_item(cls, val: str) -> tuple[str, Callable]:
        """Get item from GFS vocabulary."""
        gfs_key = cls.VOCAB[val]
        if gfs_key.split("::")[0] == "HGT":

            def mod(x: np.array) -> np.array:
                """Modify data value (if necessary)."""
                return x * 9.81

        else:

            def mod(x: np.array) -> np.array:
                """Modify data value (if necessary)."""
                return x

        return gfs_key, mod


class GEFSLexiconSel(metaclass=LexiconType):
    """Global Ensemble Forecast System 0.25 Degree Lexicon (Select variables). GEFS
    vocab specified <Data class>::<Parameter ID>::<Level/ Layer>

    Warning
    -------
    Some variables are only present for lead time greater than 0

    Note
    ----
    Additional resources:
    - https://www.nco.ncep.noaa.gov/pmb/products/gens/gec00.t00z.pgrb2s.0p25.f000.shtml
    - https://www.nco.ncep.noaa.gov/pmb/products/gens/gec00.t00z.pgrb2s.0p25.f003.shtml
    - https://www.nco.ncep.noaa.gov/pmb/docs/grib2/grib2_doc/grib2_table4-2-0-3.shtml
    """

    VOCAB = {
        "u10m": "pgrb2s::UGRD::10 m above ground",
        "v10m": "pgrb2s::VGRD::10 m above ground",
        "t2m": "pgrb2s::TMP::2 m above ground",
        "d2m": "pgrb2s::DPT::2 m above ground",
        "r2m": "pgrb2s::RH::2 m above ground",
        "sp": "pgrb2s::PRES::surface",
        "msl": "pgrb2s::PRMSL::mean sea level",  # Pressure Reduced to MSL
        "tcwv": "pgrb2s::PWAT::entire atmosphere (considered as a single layer)",
        "tp": "pgrb2s::APCP::surface",  # 3 hour acc
        "csnow": "pgrb2s::CSNOW::surface",  # 3 hour ave
        "cicep": "pgrb2s::CICEP::surface",  # 3 hour ave
        "cfrzr": "pgrb2s::CFRZR::surface",  # 3 hour ave
        "crain": "pgrb2s::CRAIN::surface",  # 3 hour ave
    }

    @classmethod
    def get_item(cls, val: str) -> tuple[str, Callable]:
        """Get item from GFS vocabulary."""
        gfs_key = cls.VOCAB[val]
        if gfs_key.split("::")[0] == "HGT":

            def mod(x: np.array) -> np.array:
                """Modify data value (if necessary)."""
                return x * 9.81

        else:

            def mod(x: np.array) -> np.array:
                """Modify data value (if necessary)."""
                return x

        return gfs_key, mod



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import os
import pathlib
import shutil
from datetime import datetime

try:
    import ecmwf.opendata as opendata
except ImportError:
    opendata = None
import numpy as np
import xarray as xr
from loguru import logger
from modulus.distributed.manager import DistributedManager
from s3fs.core import S3FileSystem
from tqdm import tqdm

from earth2studio.data.utils import datasource_cache_root, prep_data_inputs
from earth2studio.lexicon import IFSLexicon
from earth2studio.utils.type import TimeArray, VariableArray

logger.remove()
logger.add(lambda msg: tqdm.write(msg, end=""), colorize=True)


class IFS:
    """The integrated forecast system (IFS) initial state data source provided on an
    equirectangular grid. This data is part of ECMWF's open data project on AWS. This
    data source is provided on a 0.25 degree lat lon grid at 6-hour intervals for the
    most recent 4 days.

    Parameters
    ----------
    cache : bool, optional
        Cache data source on local memory, by default True
    verbose : bool, optional
        Print download progress, by default True

    Warning
    -------
    This is a remote data source and can potentially download a large amount of data
    to your local machine for large requests.

    Note
    ----
    This data source only fetches the initial state of control forecast of IFS and does
    not fetch an predicted time steps.

    Note
    ----
    Additional information on the data repository can be referenced here:

    - https://confluence.ecmwf.int/display/DAC/ECMWF+open+data%3A+real-time+forecasts
    - https://registry.opendata.aws/ecmwf-forecasts/
    - https://console.cloud.google.com/storage/browser/ecmwf-open-data/
    """

    IFS_BUCKET_NAME = "ecmwf-forecasts"
    IFS_LAT = np.linspace(90, -90, 721)
    IFS_LON = np.linspace(0, 359.75, 1440)

    def __init__(self, cache: bool = True, verbose: bool = True):
        # Optional import not installed error
        if opendata is None:
            raise ImportError(
                "ecmwf-opendata is not installed, install manually or using `pip install earth2studio[data]`"
            )

        self._cache = cache
        self._verbose = verbose
        self.client = opendata.Client(source="azure")

    def __call__(
        self,
        time: datetime | list[datetime] | TimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Function to get data.

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Timestamps to return data for (UTC).
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return. Must be in IFS lexicon.

        Returns
        -------
        xr.DataArray
            IFS weather data array
        """
        time, variable = prep_data_inputs(time, variable)

        # Create cache dir if doesnt exist
        pathlib.Path(self.cache).mkdir(parents=True, exist_ok=True)

        # Make sure input time is valid
        self._validate_time(time)

        # Fetch index file for requested time
        data_arrays = []
        for t0 in time:
            data_array = self.fetch_ifs_dataarray(t0, variable)
            data_arrays.append(data_array)

        # Delete cache if needed
        if not self._cache:
            shutil.rmtree(self.cache)

        return xr.concat(data_arrays, dim="time")

    def fetch_ifs_dataarray(
        self,
        time: datetime,
        variables: list[str],
    ) -> xr.DataArray:
        """Retrives IFS data array for given date time by fetching variable grib files
        using the ecmwf opendata package and combining grib files into a data array.

        Parameters
        ----------
        time : datetime
            Date time to fetch
        variables : list[str]
            list of atmosphric variables to fetch. Must be supported in IFS lexicon

        Returns
        -------
        xr.DataArray
            IFS data array for given date time
        """
        ifsda = xr.DataArray(
            data=np.empty((1, len(variables), len(self.IFS_LAT), len(self.IFS_LON))),
            dims=["time", "variable", "lat", "lon"],
            coords={
                "time": [time],
                "variable": variables,
                "lat": self.IFS_LAT,
                "lon": self.IFS_LON,
            },
        )

        # TODO: Add MP here, can further optimize by combining pressure levels
        # Not doing until tested.
        for i, variable in enumerate(
            tqdm(
                variables, desc=f"Fetching IFS for {time}", disable=(not self._verbose)
            )
        ):
            # Convert from Earth2Studio variable ID to GFS id and modifier
            try:
                ifs_name, modifier = IFSLexicon[variable]
            except KeyError as e:
                logger.error(f"Variable id {variable} not found in IFS lexicon")
                raise e

            variable, levtype, level = ifs_name.split("::")

            logger.debug(f"Fetching IFS grib file for variable: {variable} at {time}")
            grib_file = self._download_ifs_grib_cached(variable, levtype, level, time)
            # Open into xarray data-array
            # Provided [-180, 180], roll to [0, 360]
            da = xr.open_dataarray(
                grib_file, engine="cfgrib", backend_kwargs={"indexpath": ""}
            ).roll(longitude=-len(self.IFS_LON) // 2, roll_coords=True)
            ifsda[0, i] = modifier(da.values)

        return ifsda

    @classmethod
    def _validate_time(cls, times: list[datetime]) -> None:
        """Verify if date time is valid for IFS

        Parameters
        ----------
        times : list[datetime]
            list of date times to fetch data
        """
        for time in times:
            if not (time - datetime(1900, 1, 1)).total_seconds() % 21600 == 0:
                raise ValueError(
                    f"Requested date time {time} needs to be 6 hour interval for IFS"
                )

            if (datetime.now() - time).days > 4:
                raise ValueError(
                    f"Requested date time {time} needs to be within the past 4 days for IFS"
                )

            # if not self.available(time):
            #     raise ValueError(f"Requested date time {time} not available in IFS")

    def _download_ifs_grib_cached(
        self,
        variable: str,
        levtype: str,
        level: str | list[str],
        time: datetime,
    ) -> str:
        if isinstance(level, str):
            level = [level]

        sha = hashlib.sha256(f"{variable}_{levtype}_{'_'.join(level)}_{time}".encode())
        filename = sha.hexdigest()

        cache_path = os.path.join(self.cache, filename)

        if not pathlib.Path(cache_path).is_file():
            request = {
                "date": time,
                "type": "fc",
                "param": variable,
                "levtype": levtype,
                "step": 0,  # Would change this for forecasts
                "target": cache_path,
            }
            if levtype == "pl":  # Pressure levels
                request["levelist"] = level
            # Download
            self.client.retrieve(**request)

        return cache_path

    @property
    def cache(self) -> str:
        """Get the appropriate cache location."""
        cache_location = os.path.join(datasource_cache_root(), "ifs")
        if not self._cache:
            if not DistributedManager.is_initialized():
                DistributedManager.initialize()
            cache_location = os.path.join(
                cache_location, f"tmp_{DistributedManager().rank}"
            )
        return cache_location

    @classmethod
    def available(
        cls,
        time: datetime | np.datetime64,
    ) -> bool:
        """Checks if given date time is avaliable in the IFS AWS data store

        Parameters
        ----------
        time : datetime | np.datetime64
            Date time to access

        Returns
        -------
        bool
            If date time is avaiable
        """
        if isinstance(time, np.datetime64):  # np.datetime64 -> datetime
            _unix = np.datetime64(0, "s")
            _ds = np.timedelta64(1, "s")
            time = datetime.utcfromtimestamp((time - _unix) / _ds)

        # Offline checks
        try:
            cls._validate_time([time])
        except ValueError:
            return False

        fs = S3FileSystem(anon=True)

        file_name = f"{time.year}{time.month:0>2}{time.day:0>2}/{time.hour:0>2}z/"
        s3_uri = f"s3://{cls.IFS_BUCKET_NAME}/{file_name}"
        exists = fs.exists(s3_uri)

        return exists
